{- |
Module      : CheckingConfluence
Description :
Copyright   : (c) Jonas Schöpf, 2023
License     : GPL-3
Maintainer  : jonas.schoepf@uibk.ac.at
Stability   : stable


This module provides the main functionality to check a given LCTRS for the confluence
property.
-}
module Analysis.Confluence.CheckingConfluence (
  -- * Confluence Analysis
  analyzeConfluence,
  analyzeConfluenceSeq,

  -- * Wrapper Functions for (Non-)CR Methods
  collectMethods,
  allMethods,
  crMethods,
  nonCrMethods,
  chooseMethod,
  numberOfAllMethods,

  -- * Helper Functions
  handleError,
  isCR,
  isNonCR,
  analyzedCR,
)
where

import Analysis.Confluence.Confluence (CRResult (..))
import Analysis.Confluence.DevelopmentClosedness (
  isAlmostDevelopmentClosed,
  isAlmostDevelopmentClosedH,
  isDevelopmentClosed,
 )
import Analysis.Confluence.NewmansLemma (
  isCRbyNewmansLemma,
 )
import Analysis.Confluence.NonConfluence (
  checkNonConfluent,
 )
import Analysis.Confluence.Orthogonality (
  isOrthogonal,
  isWeaklyOrthogonal,
 )
import Analysis.Confluence.ParallelClosedness (
  isAlmostParallelClosed,
  isAlmostParallelClosedH,
  isParallelClosed,
 )
import Analysis.Confluence.StronglyClosedness (
  isStronglyClosed,
 )
import Analysis.Confluence.Toyama81 (toyama81Applies)
import Control.Monad.IO.Class (MonadIO (liftIO))
import Data.LCTRS.FIdentifier (
  FId,
 )
import Data.LCTRS.LCTRS (
  LCTRS,
  definesToSExpr,
 )
import Data.LCTRS.VIdentifier (
  VId,
 )
import Data.Maybe (fromMaybe)
import Data.Monad
import Data.SExpr
import qualified Data.Set as S
import Prettyprinter
import Rewriting.CriticalPair (CriticalPair, computeCPs)
import Rewriting.ParallelCriticalPair (computeParallelCPs)
import SimpleSMT (Solver (stop))
import Utils

----------------------------------------------------------------------------------------------------
-- confluence analysis
----------------------------------------------------------------------------------------------------

-- | confluence analysis with all methods executed asynchronously
analyzeConfluence
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => ExtSolver
  -> [Solver]
  -> Maybe [Int]
  -> Bool
  -> Int
  -> LCTRS (FId f) (VId v)
  -> FreshM CRResult
analyzeConfluence extsolver solvers methodI debugFlag solverTO lc = do
  cpsE <- preComputeCPs extsolver debugFlag solverTO lc
  case cpsE of
    Left error ->
      return $ handleError debugFlag $ Left error
    Right cps -> do
      let collectedMethods = collectMethods lc cps methodI
      let matchSolverMethod = lenSensitiveZip collectedMethods solvers
      next <- freshI
      let computations =
            [ evalStateM solver (definesToSExpr lc) debugFlag next method
            | (method, solver) <- matchSolverMethod
            ]
      jobs <- liftIO $ asyncs computations
      result <- liftIO $ waitF analyzedCR (Right MaybeConfluent) $ S.fromList jobs
      return $ handleError debugFlag result

-- | confluence analysis with all methods executed sequentially
analyzeConfluenceSeq
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => ExtSolver
  -> [Solver]
  -> Maybe [Int]
  -> Bool
  -> Int
  -> LCTRS (FId f) (VId v)
  -> FreshM CRResult
analyzeConfluenceSeq extsolver solvers methodI debugFlag solverTO lc = do
  cpsE <- preComputeCPs extsolver debugFlag solverTO lc
  case cpsE of
    Left error ->
      return $ handleError debugFlag $ Left error
    Right cps -> do
      let collectedMethods = collectMethods lc cps methodI
      let matchSolverMethod = lenSensitiveZip collectedMethods solvers
      next <- freshI
      let computations =
            [ evalStateM solver (definesToSExpr lc) debugFlag next method
            | (method, solver) <- matchSolverMethod
            ]
      wasFound <- liftIO $ findMs analyzedCR computations
      return . handleError debugFlag $
        fromMaybe (Right MaybeConfluent) wasFound

-- | function to compute critical pairs a priori
preComputeCPs
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => ExtSolver
  -> Bool
  -> Int
  -> LCTRS (FId f) (VId v)
  -> FreshM (Either Error [CriticalPair (FId f) (VId v) (VId v)])
preComputeCPs extsolver debugFlag solverTO lctrs = do
  s <- liftIO $ chooseSolver solverTO debugFlag extsolver
  val <- evalComp s (definesToSExpr lctrs) False (computeCPs lctrs)
  val <$ liftIO (stop s)

----------------------------------------------------------------------------------------------------
-- methods
----------------------------------------------------------------------------------------------------

-- | collect methods from the argument integer list
collectMethods
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> [CriticalPair (FId f) (VId v) (VId v)]
  -> Maybe [Int]
  -> [StateM CRResult]
collectMethods lc cps = maybe (allMethods lc cps) (chooseMethod lc cps)

-- | returns a list of all available (non-)confluence methods
allMethods
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> [CriticalPair (FId f) (VId v) (VId v)]
  -> [StateM CRResult]
allMethods lc cps = crMethods lc cps ++ nonCrMethods lc cps

-- | returns a list of all available confluence methods
crMethods
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> [CriticalPair (FId f) (VId v) (VId v)]
  -> [StateM CRResult]
crMethods lc cps = chooseMethod lc cps $ [1 .. 9] ++ [11 .. 12]

-- | returns a list of all available non-confluence methods
nonCrMethods
  :: (Ord v, Ord f, Pretty f, Pretty v, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> [CriticalPair (FId f) (VId v) (VId v)]
  -> [StateM CRResult]
nonCrMethods lc cps = chooseMethod lc cps [10]

-- | choose a method from a given integer
chooseMethod
  :: (Ord v, Ord f, Pretty v, Pretty f, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> [CriticalPair (FId f) (VId v) (VId v)]
  -> [Int]
  -> [StateM CRResult]
chooseMethod lctrs cps ids =
  [fInt id | id <- ids]
 where
  fInt 1 = isOrthogonal lctrs cps
  fInt 2 = isWeaklyOrthogonal lctrs cps
  fInt 3 = isStronglyClosed lctrs cps
  fInt 4 = isParallelClosed lctrs cps
  fInt 5 = isAlmostParallelClosed lctrs cps
  fInt 6 = isAlmostParallelClosedH lctrs cps
  fInt 7 = isDevelopmentClosed lctrs cps
  fInt 8 = isAlmostDevelopmentClosed lctrs cps
  fInt 9 = isAlmostDevelopmentClosedH lctrs cps
  fInt 10 = checkNonConfluent lctrs cps
  fInt 11 = isCRbyNewmansLemma lctrs cps
  fInt 12 = do
    pcps <- computeParallelCPs lctrs
    toyama81Applies lctrs cps pcps
  fInt i =
    error $
      "CheckingConfluence.hs: Unknown method \""
        ++ shows i "\" specified with \"-m\" flag, see \"-h\"."

-- | the number of all implemented confluence methods
numberOfAllMethods :: Int
numberOfAllMethods = 12

----------------------------------------------------------------------------------------------------
-- helper functions
----------------------------------------------------------------------------------------------------

-- | handle an error returned from a computation
handleError :: Bool -> Either Error CRResult -> CRResult
handleError False (Left (MayError _)) = MaybeConfluent
-- for debugging if the result is MAYBE because of an non severe error
handleError True (Left (MayError s)) = error $ "A non-severe error occured in the confluence analysis " ++ s
handleError False (Left (SevError _)) = MaybeConfluent
-- for debugging if the result is MAYBE because of an error
handleError True (Left (SevError s)) = error $ "A severe error occured in the confluence analysis " ++ s
handleError _ (Right r) = r

-- | does the given Either type consists of a confluence result
isCR :: Either Error CRResult -> Bool
isCR (Right (Confluent _)) = True
-- for debugging if the result is MAYBE because of an non severe error
-- isCR (Left (MayError e)) = error $ "A non-severe error occured in the confluence analysis " ++ e -- Debugging
isCR (Left (SevError e)) = error $ "A severe error occured in the confluence analysis " ++ e -- Debugging
isCR _ = False

-- | does the given Either type consists of a non-confluence result
isNonCR :: Either Error CRResult -> Bool
isNonCR (Right (NonConfluent _)) = True
-- for debugging if the result is MAYBE because of an non severe error
-- isNonCR (Left (MayError e)) = error $ "A non-severe error occured in the confluence analysis " ++ e -- Debugging
isNonCR (Left (SevError e)) = error $ "A severe error occured in the confluence analysis " ++ e -- Debugging
isNonCR _ = False

-- | does the given Either type consists of a (non-)confluence result
analyzedCR :: Either Error CRResult -> Bool
analyzedCR (Right (Confluent _)) = True
analyzedCR (Right (NonConfluent _)) = True
-- for debugging if the result is MAYBE because of an non severe error
-- analyzedCR (Left (MayError e)) = error $ "A non-severe error occured in the confluence analysis " ++ e -- Debugging
analyzedCR (Left (SevError e)) = error $ "A severe error occured in the confluence analysis " ++ e -- Debugging
analyzedCR _ = False
