module Analysis.Termination.CheckingTermination where

import Analysis.Termination.ConstrainedReductionOrder (rpoDP, rpoRP, rpoStrict)
import Analysis.Termination.DependencyGraph (
  approximateDPGraph,
  stronglyConnComps,
  transformDPProblemWithInfo,
 )
import Analysis.Termination.DependencyPairs (
  DPProblem (DPProblem, strictrules, weakrules),
  dpproblem,
 )
import Analysis.Termination.ReductionPair (allRhsSubtermsGroundLogicalTerms, applyReductionPair)
import Analysis.Termination.Termination (
  SNInfo (..),
  SNInfoList (SNInfoList),
  SNResult (MaybeTerminating, NonTerminating, Terminating),
  emptyPrecedence,
 )
import Analysis.Termination.ValueCriterion (valueCriterion, valueCriterionRP)
import Control.Monad.Cont (liftIO, mapAndUnzipM)
import Data.LCTRS.FIdentifier (FId)
import Data.LCTRS.LCTRS (LCTRS, definesToSExpr, getRules)
import Data.LCTRS.VIdentifier (VId)
import Data.Monad (
  Error (MayError, SevError),
  FreshM,
  StateM,
  evalStateM,
  freshI,
 )
import Data.SExpr (ToSExpr)
import qualified Data.Set as S
import Prettyprinter (Pretty (..))
import SimpleSMT (Solver)
import Utils (asyncs, lenSensitiveZip, waitF)

----------------------------------------------------------------------------------------------------
-- main functionality
----------------------------------------------------------------------------------------------------

analyzeTermination
  :: (Ord f, Ord v, ToSExpr f, ToSExpr v, Pretty f, Pretty v)
  => [Solver]
  -> Maybe [Int]
  -> Bool
  -> LCTRS (FId f) (VId v)
  -> FreshM (SNResult, SNInfoList (FId f) (VId v))
analyzeTermination solvers methodI debugFlag lctrs = do
  next <- freshI
  let collectedMethods = collectMethods methodI lctrs
  let matchSolverMethod = lenSensitiveZip collectedMethods solvers
  let computations =
        [ evalStateM solver (definesToSExpr lctrs) debugFlag next method
        | (method, solver) <- matchSolverMethod
        ]
  jobs <- liftIO $ asyncs computations
  result <- liftIO $ waitF analyzedSN (Right (MaybeTerminating, SNInfoList [None])) $ S.fromList jobs
  return $ handleError debugFlag result

collectMethods
  :: (Ord f, Ord v, ToSExpr f, ToSExpr v, Pretty f, Pretty v)
  => Maybe [Int]
  -> LCTRS (FId f) (VId v)
  -> [StateM (SNResult, SNInfoList (FId f) (VId v))]
collectMethods methodI lctrs =
  case methodI of
    Just is -> map snd . filter ((`elem` is) . fst) $ methods
    Nothing -> map snd methods
 where
  methods =
    zip
      [1 ..]
      [ terminatingByDependencyGraph lctrs
      , terminatingByRPO lctrs
      , terminatingByValueCriterion lctrs
      , terminatingByRPOwithDP lctrs
      , terminatingByReductionPair lctrs
      ]

----------------------------------------------------------------------------------------------------
-- termination methods
----------------------------------------------------------------------------------------------------

terminatingByDependencyGraph
  :: (Ord f, Ord v, ToSExpr f, ToSExpr v, Pretty f, Pretty v)
  => LCTRS (FId f) (VId v)
  -> StateM (SNResult, SNInfoList (FId f) (VId v))
terminatingByDependencyGraph lctrs = do
  dpgraph <- approximateDPGraph lctrs
  let sccs = stronglyConnComps dpgraph
  if null sccs
    then return (Terminating, SNInfoList [DpGraph (dpgraph, sccs)])
    else return (MaybeTerminating, SNInfoList [DpGraph (dpgraph, sccs)])

terminatingByValueCriterion
  :: (Ord f, Ord v, ToSExpr f, ToSExpr v, Pretty f, Pretty v)
  => LCTRS (FId f) (VId v)
  -> StateM (SNResult, SNInfoList (FId f) (VId v))
-- terminatingByValueCriterion DPProblem{..} | null strictrules = return (Terminating, SNInfoList [NoDependencyPairs])
terminatingByValueCriterion lctrs =
  let dpp = dpproblem lctrs
  in  if null $ strictrules dpp
        then return (Terminating, SNInfoList [NoDependencyPairs])
        else do
          (info, sccs) <- transformDPProblemWithInfo dpp
          (ress, infos) <- mapAndUnzipM (go []) sccs
          if all (== Terminating) ress
            then return (Terminating, SNInfoList $ DpGraph info : map (VcProjection . reverse) infos)
            else return (MaybeTerminating, SNInfoList [None])
 where
  go iterations currentDPP
    | null $ strictrules currentDPP = return (Terminating, iterations)
    | otherwise = do
        (prjs, newDPP) <- valueCriterion currentDPP
        if newDPP == currentDPP
          then return (MaybeTerminating, [])
          else case prjs of
            Nothing -> go iterations newDPP -- NOTE: should never happen
            Just ps -> go (ps : iterations) newDPP

terminatingByRPO
  :: (Ord f, Ord v, ToSExpr f, ToSExpr v, Pretty f, Pretty v)
  => LCTRS (FId f) (VId v)
  -> StateM (SNResult, SNInfoList (FId f) (VId v))
terminatingByRPO lctrs = do
  (res, info) <- rpoStrict $ getRules lctrs
  case info of
    Just i -> return (res, SNInfoList [RpoPrecedence i])
    Nothing -> return (res, SNInfoList [RpoPrecedence emptyPrecedence])

terminatingByRPOwithDP
  :: (Ord f, Ord v, ToSExpr f, ToSExpr v, Pretty f, Pretty v)
  => LCTRS (FId f) (VId v)
  -> StateM (SNResult, SNInfoList (FId f) (VId v))
terminatingByRPOwithDP lctrs =
  let dpp = dpproblem lctrs
  in  if null $ strictrules dpp
        then return (Terminating, SNInfoList [NoDependencyPairs])
        else do
          (info, sccs) <- transformDPProblemWithInfo dpp
          (ress, infos) <- mapAndUnzipM rpoDP sccs
          if all (== Terminating) ress
            then
              return
                ( Terminating
                , SNInfoList $ DpGraph info : map (maybe (RpoPrecedence emptyPrecedence) RpoPrecedence) infos
                )
            else return (MaybeTerminating, SNInfoList [None])

----------------------------------------------------------------------------------------------------
-- termination methods on reduction pairs
----------------------------------------------------------------------------------------------------

terminatingByReductionPair
  :: (Ord v, Ord f, Pretty f, ToSExpr v, ToSExpr f, Pretty v)
  => LCTRS (FId f) (VId v)
  -> StateM (SNResult, SNInfoList (FId f) (VId v))
terminatingByReductionPair lctrs = do
  let dpp = dpproblem lctrs
  if null $ strictrules dpp
    then return (Terminating, SNInfoList [NoDependencyPairs])
    else do
      (info, sccs) <- transformDPProblemWithInfo dpp
      (ress, infos) <- mapAndUnzipM (go []) sccs
      if all (== Terminating) ress
        then return (Terminating, SNInfoList $ DpGraph info : map ReductionPair infos)
        else return (MaybeTerminating, SNInfoList [None])
 where
  -- NOTE: IMPORTANT
  -- be aware of the third condition which demands that the weak
  -- part of the reduction pair orients all calculation rules weakly
  rps = [rpoRP, valueCriterionRP] -- , polynomialnterpretationsRP]
  needWeakRules dpp@DPProblem{..}
    | all allRhsSubtermsGroundLogicalTerms strictrules = DPProblem strictrules []
    | otherwise = dpp

  findApplied [] = return Nothing
  findApplied (computation : remaining) = do
    value <- computation
    case value of
      Nothing -> findApplied remaining
      Just result -> return $ Just result

  go sninfos dpp = do
    let currentDPP = needWeakRules dpp
    applied <- findApplied [applyReductionPair currentDPP rp | rp <- rps]
    case applied of
      Nothing -> return (MaybeTerminating, sninfos)
      Just (newDPP, sninfo)
        | null (strictrules newDPP) -> return (Terminating, (currentDPP, sninfo) : sninfos)
        | currentDPP == newDPP && not (null (strictrules newDPP)) ->
            return (MaybeTerminating, (currentDPP, sninfo) : sninfos)
        | otherwise -> go ((currentDPP, sninfo) : sninfos) newDPP

----------------------------------------------------------------------------------------------------
-- auxiliary methods
----------------------------------------------------------------------------------------------------

handleError :: Bool -> Either Error (SNResult, SNInfoList f v) -> (SNResult, SNInfoList f v)
handleError False (Left (MayError _)) = (MaybeTerminating, SNInfoList [None])
-- for debugging if the result is MAYBE because of an non severe error
handleError True (Left (MayError s)) = error $ "A non-severe error occured in the termination analysis " ++ s
handleError False (Left (SevError _)) = (MaybeTerminating, SNInfoList [None])
-- for debugging if the result is MAYBE because of an error
handleError True (Left (SevError s)) = error $ "A severe error occured in the termination analysis " ++ s
handleError _ (Right r) = r

analyzedSN :: Either Error (SNResult, SNInfoList f v) -> Bool
analyzedSN (Right (Terminating, _)) = True
analyzedSN (Right (NonTerminating, _)) = True
-- for debugging if the result is MAYBE because of an non severe error
-- analyzedSN (Left (MayError e)) = error $ "A non-severe error occured in the termination analysis " ++ e -- Debugging
analyzedSN (Left (SevError e)) = error $ "A severe error occured in the termination analysis " ++ e -- Debugging
analyzedSN _ = False
