module Analysis.Termination.Termination where

import qualified Analysis.Confluence.Confluence as C
import Analysis.Termination.DependencyGraph (DPGraph (..), SCC (..))
import Analysis.Termination.DependencyPairs (DPProblem)
import Data.LCTRS.Rule (Rule, prettyRule)
import Data.LCTRS.Sort (Attr (..), Sort (..), Sorted (sort))
import qualified Data.Map.Strict as M
import Data.SExpr (ToSExpr (toSExpr))
import Data.SMT (boolSort, intSort, isBVSort, realSort)
import Prettyprinter (Doc, Pretty (pretty), encloseSep, indent, vsep, (<+>))
import SimpleSMT (SExpr)
import qualified SimpleSMT as SMT

----------------------------------------------------------------------------------------------------
-- comparison and zero elements of theory sorts
----------------------------------------------------------------------------------------------------

getCompAndZeroOfSort
  :: Sort
  -> Maybe
      ( SExpr -> SExpr -> SExpr
      , SExpr -> SExpr -> SExpr
      , SExpr
      )
getCompAndZeroOfSort s | s == boolSort = Nothing -- return (SMT.gt, SMT.geq, SMT.const "false")
getCompAndZeroOfSort s | s == intSort = (SMT.gt,SMT.geq,) <$> zeroSExprOfSort s
getCompAndZeroOfSort s | s == realSort = (SMT.gt,SMT.geq,) <$> zeroSExprOfSort s
getCompAndZeroOfSort s
  | isBVSort s =
      (\s1 s2 -> SMT.fun "bvugt" [s1, s2],\s1 s2 -> SMT.fun "bvuge" [s1, s2],)
        <$> zeroSExprOfSort s
getCompAndZeroOfSort _ = Nothing

zeroSExprOfSort :: Sort -> Maybe SExpr
zeroSExprOfSort s | s == intSort = return $ SMT.const "0"
zeroSExprOfSort s | s == realSort = return $ SMT.const "0.0"
zeroSExprOfSort s | isBVSort s = correctZeroBVConst s
zeroSExprOfSort _ = Nothing

correctZeroBVConst :: Sort -> Maybe SExpr
correctZeroBVConst (AttrSort "BitVec" (AttrInt i)) = return $ SMT.const $ '#' : 'b' : zeroes
 where
  zeroes = replicate i '0'
correctZeroBVConst _ = Nothing

----------------------------------------------------------------------------------------------------
----------------------------------------------------------------------------------------------------

-- precedence from least to greatest
-- the inner set means equality for those symbols
newtype Precedence f = Prec [[f]]

emptyPrecedence :: Precedence f
emptyPrecedence = Prec []

data SNResult = Terminating | NonTerminating | MaybeTerminating deriving (Eq)

newtype SNInfoList f v = SNInfoList [SNInfo f v]

instance (Pretty f, Pretty v, Ord f) => Pretty (SNInfoList f v) where
  pretty (SNInfoList sninfos) = vsep $ map pretty sninfos

data SNInfo f v
  = RpoPrecedence (Precedence f)
  | VcProjection [([Rule f v], [(f, Int)])]
  | NoDependencyPairs
  | DpGraph (DPGraph f v, [SCC f v])
  | PolyInterpretation (M.Map f ([Int], Int)) -- [String]
  | ReductionPair [(DPProblem f v, SNInfo f v)]
  | None

listSet :: [Doc ann] -> Doc ann
listSet = encloseSep "{" "}" ", "

instance (Ord f, Pretty f, Pretty v) => Pretty (SNInfo f v) where
  pretty (RpoPrecedence (Prec [])) = "RPO with empty precedence."
  pretty (RpoPrecedence (Prec prec)) =
    vsep
      [ "RPO with precedence:"
      , indent 2 $
          encloseSep "" "" " > " $
            map (encloseSep "{" "}" " " . map pretty) $
              reverse prec
      ]
  pretty (VcProjection []) = "Value criterion with empty projection."
  pretty (VcProjection proj) =
    vsep
      [ "Value criterion after" <+> pretty (length proj) <+> "iteration(s) with the projections:"
      , ( indent 2 $
            vsep $
              zipWith
                ( \i (rules, projs) ->
                    vsep
                      [ "Iteration" <+> pretty i
                          <> ":"
                      , indent 2 "projection(s):"
                      , indent 4 $
                          encloseSep
                            ""
                            ""
                            ", "
                            (map (\(f, param) -> "v(" <> pretty f <> ")" <+> "=" <+> pretty (succ param)) projs)
                      , indent 2 "removing the rule(s):"
                      , indent 4 $ listSet (map prettyRule rules)
                      ]
                )
                [1 :: Int ..]
                proj
        )
      ]
  pretty NoDependencyPairs = "No Dependency Pairs."
  pretty (DpGraph (dpgraph, sccs)) =
    vsep
      [ pretty dpgraph
      , "with" <+> pretty (length sccs) <+> "SCC(s)"
      , indent 2 $ vsep $ map pretty sccs
      ]
  pretty (PolyInterpretation m) =
    vsep
      [ "Polynomial Interpretations:"
      , indent 2 $
          vsep
            [ pretty f <> "(" <> args <> "):" <+> poly <+> "+" <+> pretty c
            | (f, (cs, c)) <- M.toList m
            , let ics = zip [1 :: Int ..] cs
            , let vars = map (("x" <>) . pretty . fst) ics
            , let args = encloseSep "(" ")" ", " vars
            , let poly = encloseSep "" "" " + " $ zipWith (\c' v -> pretty c' <+> "*" <+> v) cs vars
            ]
      ]
  pretty (ReductionPair sninfos) =
    vsep
      [ "Reduction Pair with dependency pairs:"
      , indent 2 $
          vsep $
            map
              ( \(dpp, sninfo) ->
                  indent 2 $
                    vsep [pretty dpp, pretty sninfo]
              )
              sninfos
      ]
  pretty None = "No termination info given."

snresultToAnswer :: (SNResult, SNInfoList f v) -> C.Answer
snresultToAnswer (Terminating, _) = C.Yes
snresultToAnswer (NonTerminating, _) = C.No
snresultToAnswer (MaybeTerminating, _) = C.Maybe

-- printSnInfo ::(SNResult, SNInfo f) -> String

cutoffZeroSExpr :: (Sorted a, ToSExpr a) => a -> Maybe SExpr
cutoffZeroSExpr expr = do
  (gt, _, zero) <- getCompAndZeroOfSort (sort expr)
  return $ SMT.ite (gt sexpr zero) sexpr zero
 where
  sexpr = toSExpr expr
