{- |
Module      : CriticalPair
Description :
Copyright   : (c) Jonas Schöpf, 2023
License     : GPL-3
Maintainer  : jonas.schoepf@uibk.ac.at
Stability   : stable


This module provides an implementation for constrained critical pairs and how to
compute them.
-}
module Rewriting.CriticalPair where

import Control.Monad (
  filterM,
  replicateM,
 )
import Data.LCTRS.FIdentifier
import Data.LCTRS.Guard
import Data.LCTRS.LCTRS hiding (rename)
import Data.LCTRS.Position (Pos, epsilon, position)
import Data.LCTRS.Rule hiding (guard)
import qualified Data.LCTRS.Rule as R
import Data.LCTRS.Sort
import Data.LCTRS.Term
import qualified Data.LCTRS.Term as T
import Data.LCTRS.VIdentifier
import qualified Data.Map.Strict as M
import Data.Maybe (fromJust)
import Data.Monad (
  StateM,
  freshI,
 )
import qualified Data.Rewriting.CriticalPair.Ops as CP
import qualified Data.Rewriting.CriticalPair.Type as CP
import Data.Rewriting.Substitution.Type (toMap)
import qualified Data.Rewriting.Substitution.Type as Sub
import Data.SExpr
import Data.SMT (boolSort)
import qualified Data.Set as S
import Prettyprinter
import Rewriting.Renaming (Renaming, eitherToRenaming)
import Rewriting.SMT (
  satGuard,
  smtResultCheck,
 )
import qualified Rewriting.Substitution as MySub
import Utils (parMap)

----------------------------------------------------------------------------------------------------
-- critical pair definition
----------------------------------------------------------------------------------------------------

data CriticalPair f v v' = CriticalPair
  { innerRule :: Rule f (Renaming v v')
  , outerRule :: Rule f (Renaming v v')
  , top :: Term f (Renaming v v')
  , inner :: Term f (Renaming v v')
  , outer :: Term f (Renaming v v')
  , constraint :: Guard f (Renaming v v')
  , pos :: Pos
  , subst :: MySub.Subst f (Renaming v v')
  }

critp
  :: (Ord v, Ord v', Ord f)
  => M.Map f T.FunType
  -> Rule f (Either v v')
  -> Rule f (Either v v')
  -> CP.CP f v v'
  -> Guard f (Either v v')
  -> CriticalPair f v v'
critp types innerR outerR cp cons = CriticalPair r1 r2 t i o g p s
 where
  revert = fromJust . T.fromTTerm types

  r1 = rename eitherToRenaming innerR
  r2 = rename eitherToRenaming outerR
  t = mapVars eitherToRenaming . revert $ CP.top cp
  i = mapVars eitherToRenaming . revert $ CP.left cp
  o = mapVars eitherToRenaming . revert $ CP.right cp
  g = mapGuard (mapVars eitherToRenaming) cons
  p = position $ CP.leftPos cp
  s =
    let subs = M.toList . toMap $ CP.subst cp
    in  MySub.fromMap . M.fromList $
          [(eitherToRenaming v, mapVars eitherToRenaming (revert t)) | (v, t) <- subs]

----------------------------------------------------------------------------------------------------
-- computing the set of critical pairs
----------------------------------------------------------------------------------------------------

computeCPs
  :: (Ord v, Ord f, Pretty v, Pretty f, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> StateM [CriticalPair (FId f) (VId v) (VId v)]
computeCPs l = (++) <$> computeRuleCPs l <*> computeCalcCPs l

-- critical pairs between standard rules
computeRuleCPs
  :: (Ord v, Ord f, Pretty v, Pretty f, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> StateM [CriticalPair (FId f) (VId v) (VId v)]
computeRuleCPs l =
  -- not sure if I this parallelism should stay
  filterM criticalPairProperties $ concat $ parMap 10 cpsFromRule rs
 where
  rs = getRules l

  cpsFromRule ruleO =
    [ critp (mO <> mI) renRuleI renRuleO cp g
    | let (outer, mO) = toTRule ruleO
    , ruleI <- rs
    , let (inner, mI) = toTRule ruleI
    , cp <- CP.cps [inner] [outer]
    , let sub = MySub.fromMap . fmap (fromJust . T.fromTTerm (mO <> mI)) . Sub.toMap $ CP.subst cp
    , let renRuleO = rename Right ruleO
    , let renRuleI = rename Left ruleI
    , all (check . MySub.apply sub . T.var) $
        R.lvar renRuleO
          `S.union` R.lvar renRuleI
    , let g' = mapGuard (MySub.apply sub) (conjGuards (R.guard renRuleI) (R.guard renRuleO))
    , let innerEVars = extractEVars Left ruleI sub
    , let outerEVars = extractEVars Right ruleO sub
    , let g = createGuard $ (termsGuard g' ++) $ innerEVars ++ outerEVars
    ]

  check t = T.isVar t || T.isValue t

-- critical pairs between standard rules and calculation rules
computeCalcCPs
  :: (Ord v, Ord f, Pretty v, Pretty f, ToSExpr f, ToSExpr v)
  => LCTRS (FId f) (VId v)
  -> StateM [CriticalPair (FId f) (VId v) (VId v)]
computeCalcCPs lctrs = do
  -- not sure if I this parallelism should stay
  matchedCalcs <-
    sequence [(rule,) <$> calcRules | rule <- getRules lctrs, let calcRules = symRules rule]
  let cps = parMap 10 (uncurry cpsFromRule) matchedCalcs
  filterM criticalPairProperties $ concat cps
 where
  cpsFromRule ruleO symRules =
    [ critp (mO <> mI) renRuleI renRuleO cp g
    | let (outer, mO) = toTRule ruleO
    , ruleI <- symRules
    , let (inner, mI) = toTRule ruleI
    , cp <- CP.cps [inner] [outer]
    , let sub = MySub.fromMap . fmap (fromJust . T.fromTTerm (mO <> mI)) . Sub.toMap $ CP.subst cp
    , let renRuleO = rename Right ruleO
    , let renRuleI = rename Left ruleI
    , all (check . MySub.apply sub . T.var) $
        R.lvar renRuleO
          `S.union` R.lvar renRuleI
    , let g' = mapGuard (MySub.apply sub) (conjGuards (R.guard renRuleI) (R.guard renRuleO))
    , let innerEVars = extractEVars Left ruleI sub
    , let outerEVars = extractEVars Right ruleO sub
    , let g = createGuard $ (termsGuard g' ++) $ innerEVars ++ outerEVars
    ]

  check t = T.isVar t || T.isValue t

  syms = S.toList $ getTheoryFuns lctrs

  symRules r = do
    ss <-
      mapM
        ( \f -> do
            let (inSs, outS) = (getFIdInSort f, sort f)
            let a = length inSs
            is <- zip inSs <$> replicateM a freshI
            rhs <- var . freshV outS <$> freshI
            return (f, T.theoryFun f (map (\(s, i) -> var $ freshV s i) is), rhs)
        )
        $ filter (not . null . getFIdInSort) syms
    return $
      [ R.createRule
        lhs
        rhs
        (createGuard [T.eq sa rhs lhs])
      | (f, lhs, rhs) <- ss
      , f `elem` R.funs r -- NOTE: optimization to reduce number of CPs
      , let sa = sortAnnotation [sort rhs, sort lhs] boolSort
      ]

extractEVars
  :: (Ord v)
  => (VId v -> Either (VId v) (VId v))
  -> Rule (FId f) (VId v)
  -> MySub.Subst (FId f) (Either (VId v) (VId v))
  -> [Term (FId f) (Either (VId v) (VId v))]
extractEVars ren rule sub =
  [ T.eq sa v v
  | n <-
      S.toList
        (extraVars rule)
  , let v = MySub.apply sub . mapVars ren $ T.var n
  , let sa = sortAnnotation [sort n, sort n] boolSort
  ]

-- create calculation rules for a given LCTRS
createCalcs
  :: LCTRS (FId f) v
  -> (Sort -> Int -> v)
  -> StateM [Rule (FId f) v]
createCalcs lctrs fresh = symRules
 where
  syms = S.toList $ getTheoryFuns lctrs

  symRules = do
    mapM
      ( \f -> do
          let (inSs, outS) = (getFIdInSort f, sort f)
          let a = length inSs
          is <- zip inSs <$> replicateM a freshI
          let lhs = T.theoryFun f [var $ fresh s i | (s, i) <- is]
          rhs <- var . fresh outS <$> freshI
          let sa = sortAnnotation [outS, outS] boolSort
          let g = createGuard [T.eq sa rhs lhs]
          return $ R.createRule lhs rhs g
      )
      $ filter (not . null . getFIdInSort) syms

criticalPairProperties
  :: (Ord v, Ord f, Pretty v, Pretty f, ToSExpr f, ToSExpr v)
  => CriticalPair (FId f) (VId v) (VId v)
  -> StateM Bool
criticalPairProperties cp@(CriticalPair{..}) = do
  satM <- satGuard constraint
  sat <- smtResultCheck satM
  return $
    respVars sub lvars
      && sat
      && ( not (isOverlay cp)
            || not (R.isVariantOf innerRule outerRule)
            || not
              ( S.fromList (T.vars (R.rhs innerRule))
                  `S.isSubsetOf` S.fromList (T.vars (R.lhs innerRule))
              )
         )
 where
  lvars = R.lvar innerRule `S.union` R.lvar outerRule
  sub = subst

----------------------------------------------------------------------------------------------------
-- auxiliary functions
----------------------------------------------------------------------------------------------------

isOverlay :: CriticalPair f v v' -> Bool
isOverlay CriticalPair{..} = pos == epsilon

respVars
  :: (Eq f, Ord v)
  => MySub.Subst (FId f) v
  -> S.Set v
  -> Bool
respVars sub = all (check . MySub.apply sub . T.var)
 where
  check t = T.isVar t || T.isValue t || T.isConst t

prettyCriticalPair :: (Pretty f, Pretty v, Pretty v') => CriticalPair f v v' -> Doc ann
prettyCriticalPair CriticalPair{..} =
  "Critical Pair"
    <> line
    <> indent
      2
      ( vsep
          [ nest 2 $ vsep ["Inner Rule", prettyRule innerRule]
          , nest 2 $ vsep ["Outer Rule", prettyRule outerRule]
          , nest 2 $ vsep ["Pair", prettyPair]
          ]
      )
 where
  prettyPair
    | isTopGuard constraint = pretty inner <+> "≈" <+> pretty outer
    | otherwise = pretty inner <+> "≈" <+> pretty outer <+> prettyGuard constraint

prettyCriticalPairs :: (Pretty f, Pretty v, Pretty v') => [CriticalPair f v v'] -> Doc ann
prettyCriticalPairs = vsep . map prettyCriticalPair

prettyCriticalPairFId :: (Pretty f, Pretty v, Pretty v') => CriticalPair (FId f) v v' -> Doc ann
prettyCriticalPairFId CriticalPair{..} =
  "Critical Pair"
    <> line
    <> indent
      2
      ( vsep
          [ nest 2 $ vsep ["Inner Rule", prettyRuleFId innerRule]
          , nest 2 $ vsep ["Outer Rule", prettyRuleFId outerRule]
          , nest 2 $ vsep ["Pair", prettyPair]
          ]
      )
 where
  prettyPair
    | isTopGuard constraint = prettyTermFId inner <+> "≈" <+> prettyTermFId outer
    | otherwise = prettyTermFId inner <+> "≈" <+> prettyTermFId outer <+> prettyGuardFId constraint

prettyCriticalPairsFId :: (Pretty f, Pretty v, Pretty v') => [CriticalPair (FId f) v v'] -> Doc ann
prettyCriticalPairsFId = vsep . map prettyCriticalPairFId
