{- |
Module      : Unification
Description :
Copyright   : (c) Jonas Schöpf, 2023
License     : GPL-3
Maintainer  : jonas.schoepf@uibk.ac.at
Stability   : stable


This module provides function to extract a valid substitution for a matching or
unification problem.
-}
module Rewriting.Unification (
  solveMatchingProblem,
  solveUnificationProblem,
  UnifPTuple (..),
  UnifProblem (..),
) where

----------------------------------------------------------------------------------------------------
-- imports
----------------------------------------------------------------------------------------------------

import Control.Monad (zipWithM)
import Control.Monad.State (StateT (runStateT), gets, modify)
import Control.Monad.Trans.Class (lift)
import Control.Monad.Union
import Data.Containers.ListUtils (nubOrd)
import Data.LCTRS.Term (Term)
import qualified Data.LCTRS.Term as T
import qualified Data.Map.Strict as M
import Data.Maybe (catMaybes, fromJust, isJust)
import qualified Data.Union as DU
import Rewriting.Substitution (Subst, fromMap)

----------------------------------------------------------------------------------------------------
-- types
----------------------------------------------------------------------------------------------------

type UnifM f v = StateT (UnifSt f v) (UnionM (Term f v))

newtype UnifSt f v = UnifSt
  { termNode :: M.Map (Term f v) Node
  }

newtype UnifPTuple f v = UnifPTuple {unifPTuple :: (Term f v, Term f v)}

newtype UnifProblem f v = UnifProblem
  { unifProblem :: [UnifPTuple f v]
  }

emptyUnifSt :: UnifSt f v
emptyUnifSt = UnifSt M.empty

evalUnifM :: UnifSt f v -> UnifM f v a -> (Union (Term f v), (a, UnifSt f v))
evalUnifM state computation = run' $ runStateT computation state

----------------------------------------------------------------------------------------------------
-- solve matching problem
----------------------------------------------------------------------------------------------------

solveMatchingProblem
  :: (Ord v, Ord f) => UnifProblem f v -> Maybe (Subst f v)
solveMatchingProblem = solveProblem True

----------------------------------------------------------------------------------------------------
-- solve unification problem
----------------------------------------------------------------------------------------------------

solveUnificationProblem
  :: (Ord v, Ord f) => UnifProblem f v -> Maybe (Subst f v)
solveUnificationProblem = solveProblem False

----------------------------------------------------------------------------------------------------
-- solving unification problem using the union find data structure
----------------------------------------------------------------------------------------------------

-- use Union find to solve the unification/matching problem
solveProblem
  :: (Ord v, Ord f) => Bool -> UnifProblem f v -> Maybe (Subst f v)
solveProblem isMatching UnifProblem{..}
  | not $ and $ catMaybes suceeded = Nothing
  | otherwise =
      return
        . fromMap -- make substitution
        $ M.fromList -- make map
          [(var, fromJust assigned) | var <- allVars, let assigned = varToTerm $ T.var var, isJust assigned]
 where
  termToNode s = M.lookup s (termNode state)
  varToTerm term = snd . DU.lookup union <$> termToNode term

  varsPT UnifPTuple{..} = let (f, s) = unifPTuple in T.vars f ++ T.vars s
  allVars = nubOrd $ concatMap varsPT unifProblem

  (union, (suceeded, state)) = evalUnifM emptyUnifSt $ mapM unifP unifProblem
  unifP UnifPTuple{..} = uncurry (solveProblemTerm isMatching) unifPTuple

solveProblemTerm
  :: (Ord f, Ord v)
  => Bool
  -> Term f v
  -> Term f v
  -> UnifM f v (Maybe Bool)
solveProblemTerm _ v@(T.Var n) term | n `notElem` T.vars term = do
  n1 <- nodeOfTerm v
  n2 <- nodeOfTerm term
  mergeN n1 n2
solveProblemTerm False term v@(T.Var n) | n `notElem` T.vars term = do
  -- no complete unification, but only matching if boolean is True
  n1 <- nodeOfTerm v
  n2 <- nodeOfTerm term
  mergeN n1 n2
solveProblemTerm isMatching (T.Fun typF f fargs) (T.Fun typG g gargs)
  | f == g && typF == typG =
      Just . and . catMaybes <$> zipWithM (solveProblemTerm isMatching) fargs gargs
solveProblemTerm _ _ _ = return $ Just False

nodeOfTerm :: (Ord v, Ord f) => Term f v -> UnifM f v Node
nodeOfTerm term = do
  map <- gets termNode
  case M.lookup term map of
    Nothing -> do
      node <- new term
      modify $ \st -> st{termNode = M.insert term node map}
      return node
    Just n -> return n

mergeN :: (Eq v, Eq f) => Node -> Node -> UnifM f v (Maybe Bool)
mergeN n1 n2 = lift $ merge checkSort n1 n2
 where
  checkSort :: (Eq v, Eq f) => Term f v -> Term f v -> (Term f v, Bool)
  checkSort term term'
    | T.isVar term' = (term, True)
    | T.isVar term = (term', True)
    | otherwise = (term, False)
